/* Initng, a next generation sysvinit replacement.
 * Copyright (C) 2005 neuron <neuron@hollowtube.mine.nu>
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General
 * Public License along with this library; if not, write to the
 * Free Software Foundation, Inc., 51 Franklin Street, Fifth Floor,
 * Boston, MA  02110-1301, USA.
 */

#include "initng.h"


#include <stdio.h>
#include <stdlib.h>                         /* free() exit() */
#include <string.h>
#include <assert.h>

#include "initng_handler.h"
#include "initng_global.h"
#include "initng_plugin_hook.h"
#include "initng_common.h"
#include "initng_toolbox.h"
#include "initng_static_data_id.h"
#include "initng_static_states.h"

#include "initng_depend.h"

static int dep_on(active_db_h * service, active_db_h * check);


/*
 * initng_depend:
 *  Will check with all plug-ins and return TRUE
 *  if service depends on check.
 */
int initng_depend(active_db_h * service, active_db_h * check)
{
    s_call *current, *s = NULL;
    int result = FALSE;

    assert(service);
    assert(check);

    /* it can never depend on itself */
    if (service == check)
        return (FALSE);

    /* run the local static dep check */
    if (dep_on(service, check) == TRUE)
        return (TRUE);

    /* run the global plugin dep check */
    while_list_safe(current, &g.DEP_ON, s)
    {
        if ((result = (*current->c.dep_on_check) (service, check) == TRUE))
            return (TRUE);
    }

    return (FALSE);
}

/*
 * A deeper deep-find.
 * Logic, we wanna make sure that depend check in a deeper level.
 * if daemon/smbd -> daemon/samba -> system/checkroot -> system/initial.
 * So should initng_depend_deep(daemon/smbd, system/initial) == TRUE
 *
 * Summary, does service depends on check?
 */
static int initng_depend_deep2(active_db_h * service, active_db_h * check,
                               int *stack);
int initng_depend_deep(active_db_h * service, active_db_h * check)
{
    int stack = 0;

    return (initng_depend_deep2(service, check, &stack));
}

static int initng_depend_deep2(active_db_h * service, active_db_h * check,
                               int *stack)
{
    active_db_h *current = NULL;
    int result = FALSE;

    if (current == service)
        return (FALSE);

    /* if service depends on check, it also dep_on_deep's on check */
    /* this serves as an exit from the recursion */
    if (initng_depend(service, check))
        return (TRUE);

    /* in case there is a circular dependency, break after 10 levels of recursion */
    (*stack)++;
    if (*stack > 10)
        return (FALSE);

    /* loop over all services, if service depends on current, recursively check if
     * current may depend (deep) on check */
    while_active_db(current)
    {
        if (initng_depend(service, current))
        {
            if ((result = initng_depend_deep2(current, check, stack)))
                break;
        }
    }
    return result;
}



/* standard dep check */
static int dep_on(active_db_h * service, active_db_h * check)
{
    const char *string = NULL;

    assert(service);
    assert(check);

    while ((string =
            initng_active_db_get_next_string(&REQUIRE, service, string)))
    {
        if (strcmp(string, check->name) == 0)
            return (TRUE);
    }
    string = NULL;
    while ((string =
            initng_active_db_get_next_string(&NEED, service, string)))
    {
        if (strcmp(string, check->name) == 0)
            return (TRUE);
    }
    string = NULL;
    while ((string = initng_active_db_get_next_string(&USE, service, string)))
    {
        if (strcmp(string, check->name) == 0)
            return (TRUE);
    }
    return (FALSE);
}

int initng_depend_stop_deps(active_db_h * service)
{
    active_db_h *current = NULL;
    active_db_h *safe = NULL;

    /* also stop all service depending on service_to_stop */
    while_active_db_safe(current, safe)
    {
        /* Dont mind stop itself */
        if (current == service)
            continue;

        /* if current depends on the one we are stopping */
        if (initng_depend_deep(current, service) == TRUE)
            initng_handler_stop_service(current);
    }

    return (TRUE);
}


/*
 * This is a final check, before a daemon or service can be marked as UP.
 */
int initng_depend_up_check(active_db_h * service)
{
    s_call *current, *s = NULL;
    int ret;

    /* run the global plugin dep check */
    while_list_safe(current, &g.UP_MET, s)
    {

        if ((ret = (*current->c.up_met) (service)) < TRUE)
            return (ret);

    }

    /* Return happily */
    return (TRUE);

}

/*
 * This will check with plug-ins if dependencies for start this is met.
 * If this returns FALSE deps are not met yet, try later.
 * If this returns FAIL deps wont ever be met, so stop trying.
 */
int initng_depend_start_dep_met(active_db_h * service)
{
    s_call *current, *s = NULL;
    active_db_h *currentA = NULL;
    int ret;
    int i = 0;


    /* only do this check if it has not been done before */
    if (service->depend_cache == 0)
    {

        /*
         *    Check so all deps, that needs service, is up.
         */
        while_active_db(currentA)
        {
            if (currentA == service)
                continue;

            /* if current is not one that it depends on continue */
            if (initng_depend(service, currentA) == FALSE)
                continue;

            /* if service dep on is starting, wait a bit */
            if (IS_STARTING(currentA))
                return (FALSE);

            /* if service failed, return that */
            if (IS_FAILED(currentA))
                return (FAIL);

            /* if its not starting or up, return FAIL */
            if (!IS_UP(currentA))
            {
                initng_handler_start_service(currentA);
                return (FALSE);
            }
        }

        service->depend_cache++;
    }


    /* run the global plugin dep check */
    while_list_safe(current, &g.START_DEP_MET, s)
    {
        /* This is check number i */
        i++;

        /* Only run this check if it has not been run before */
        if (service->depend_cache > i)
            continue;

        if ((ret = (*current->c.start_dep_met) (service)) < TRUE)
            return (ret);

        /* if this succeds, count up one, so test wont be run again */
        service->depend_cache++;
    }

    service->depend_cache = 0;
    return (TRUE);
}

/*
 * This will check with plug-ins if dependencies for stop is met.
 * If this returns FALSE deps for stopping are not met, try again later.
 * If this returns FAIL stop deps, wont EVER be met, stop trying.
 */
int initng_depend_stop_dep_met(active_db_h * service)
{
    s_call *current, *s = NULL;
    active_db_h *currentA = NULL;
    int ret;
    int i = 0;


    /* only do this check if it has not been done before */
    if (service->depend_cache == 0)
    {

        /*
         *    Check so all deps, that needs service, is down.
         *    if there are services depending on this one still running, return false and still try
         */
        while_active_db(currentA)
        {
            if (currentA == service)
                continue;

            /* Does service depends on current ?? */
            if (initng_depend(currentA, service) == FALSE)
                continue;

            /* if its done, this is perfect */
            if (IS_DOWN(currentA))
                continue;

            /* If the dep is failed, continue */
            if (IS_FAILED(currentA))
                continue;

            /* else RETURN */
            D_("still waiting for %s state %s\n", currentA->name,
               currentA->current_state->state_name);

            /* if its still up */
            if (IS_UP(currentA))
            {
                /* stop the service again, youst to make sure */
                initng_handler_stop_service(currentA);
            }
            return (FALSE);
        }
        service->depend_cache++;
    }

    /* run the global plugin dep check */
    while_list_safe(current, &g.STOP_DEP_MET, s)
    {
        /* This is check number i */
        i++;

        /* This counts how many test thats have succeded before */
        if (service->depend_cache > i)
            continue;

        if ((ret = (*current->c.stop_dep_met) (service)) != TRUE)
            return (FALSE);

        /* count this check so it wont be run again */
        service->depend_cache++;
    }

    service->depend_cache = 0;
    return (TRUE);
}

/*
 * Start all deps, required or needed.
 * If a required deps failed to start, this will return FALSE
 */
int initng_depend_start_deps(active_db_h * service)
{
    const char *tmp = NULL;

    assert(service);
    assert(service->name);

    while ((tmp = initng_active_db_get_next_string(&REQUIRE, service, tmp)))
    {
        /* look if it exits already */
        if (initng_active_db_find_by_name(tmp))
        {
            D_("No need to LOAD %s, it is already loaded!\n", tmp);
            continue;
        }

        /* OKAY, start the service */

        if (!initng_handler_start_new_service_named(tmp))
        {
            F_("%s required dep \"%s\" could not start!\n", service->name,
               tmp);
            initng_handler_stop_service(service);
            return (FALSE);
        }
    }

    while ((tmp = initng_active_db_get_next_string(&NEED, service, tmp)))
    {
        /* look if it exits already */
        if (initng_active_db_find_by_name(tmp))
        {
            D_("No need to LOAD %s, it is already loaded!\n", tmp);
            continue;
        }

        /* OKAY, start the service */

        if (!initng_handler_start_new_service_named(tmp))
        {
            W_("service \"%s\" needs service \"%s\", that could not be found!\n", service->name, tmp);
        }
    }


    return (TRUE);
}
